; Program flow:
; 1. we are loaded at 0x7C00
; 2. init and store drive ID
; 3. do a check on MBR parition table for GPT protective record
; 4. do a check for LBA capabilities (extended read/write)
; 5. read GPT header
; 6. do a check on GPT signature
; 7. do a check on GPT partition count (should be more than or equal to 1)
; 8. read first 4 entries from GPT partition array
; 9. do a check on each partition entry for GUID of BBP
; 10. store BBP partition's start LBA
; 10. calculate and store BBP partition's lenght
; 11. relocate rest of the MBR (just the part that loads data from BBP) to 0x0700
; 12. read BBP from the disk and store in memory starting at 0x7C00
; 13. jump to BBP (0x7C00)
;
; memory map at the start:
; 0x0600 - 0x7BFF - free
; 0x7C00 - 0x7DFF - MBR loaded by BIOS
; 0x7E00 - 0x0007FFFF - free
;
; memory map before the jump to BBP:
; 0x0600 - Disk address packet (DAP) location (2x qword)
; 0x0620 - disk ID (word?, received from BIOS)
; 0x0640 - error message code
; 0x0700 - MBR bootstrap after relocation (curently it looks like just 16 bytes)
; 0x7C00 - 0x0007FFFF - Bootloader code loaded by MBR bootstrap (read buffer location in memory)

; Code location constants
%define ORG_POS			0x7C00					; Initial MBR position in memory (where BIOS loads it)
%define RELOC_POS		0x0700					; Relocation position (where we will copy neccessary MBR code to chainload bootloader)

; MBR constants
%define MBR_PART_OFF	0x01BE					; Partition table offset from the beginning of MBR

; Vital data storage locations
%define DRIVE_ID_POS	0x0620					; Address in memory where we'll store current drive ID
%define ERR_MSG_POS		0x0640					; Address in memory where we'll store error messags

; Disk read constants
%define EXT_DAP_POS		0x0600					; Address in memory where we'll store Disk address packet (DAP)
%define EXT_BUF_POS		0x0800					; -"- read buffer

; GPT constants
%define PART_COUNT_OFF	0x0050					; Partition count entry offset in GPT header
%define PART_ARR_OFF	0x0200					; Partition table offset from GPT header position
%define PART_LBA_START	0x0420					; Partition LBA start offset
%define PART_LBA_END	0x0428					; Partition LBA end offset

; Remember in NASM it's:
; instruction destination, source

[org ORG_POS]									; set origin at 0x7C00
[bits 16]										; tell NASM that we are working in 16bit environment

start:											; Start MBR procedures
	xor ax, ax									; clear ax (is this some kind of asm developers way of seting value to 0? why can't we use
												; mov ax, word 0x0000 instead? is this faster or cleaner?)
	mov ss, ax									; set stack segment to zero (is it, i'm dumb in assembly?)
	mov sp, ORG_POS								; set stack pointer to the begining of MBR location in memory
	mov es, ax									; zero-out extra segment
	mov ds, ax									; zero-out data segment
	mov [DRIVE_ID_POS], dl						; store drive ID into memory
	
test_mbr_part:									; Test weather we are a protective MBR, we certanly don't want to be a hybrid MBR
	cmp [ORG_POS + MBR_PART_OFF + 4], byte 0xEE	; check weather we have GPT partitioning (0xEE - GPT protective partition)
	je test_ext									; continiue if we're a GPT (classical MBR bootloaders can be found elswere) 
	mov [ERR_MSG_POS], byte 0x31				; store error message
	jmp print_fault								; print a message and hang

test_ext:										; Check if drive supports extended read
	mov ah, 0x41								; command - check for extensions
	mov bx, 0x55AA								; some magic test value
	mov dl, [DRIVE_ID_POS]						; set drive ID
	int 0x13									; call disk interrupt
	;jc $										; hang if extensions are not supported
	cmp bx, 0xAA55								; extra test for return value from extension test
	je reset_disk								; conitniue if the return value is the one we expected
	mov [ERR_MSG_POS], byte 0x32				; store error message
	jmp print_fault								; print a message and hang

reset_disk:										; Reset current disk
	mov ah, 0x00								; command - reset disk
	mov dl, [DRIVE_ID_POS]						; set drive ID
	int 0x13									; call disk interrupt
	jnc read_gpt								; continiue reset went OK
	mov [ERR_MSG_POS], byte 0x33				; store error message
	jmp print_fault								; print a message and hang

read_gpt:										; Read GPT header
	mov [EXT_DAP_POS], byte 0x10				; set DAP packet size
	mov [EXT_DAP_POS + 1], byte 0x00			; set reserved value to 0
	mov [EXT_DAP_POS + 2], word 0x0001			; set number of blocks to transfer
	mov [EXT_DAP_POS + 4], word EXT_BUF_POS		; set read buffer position in memory (low word)
	mov [EXT_DAP_POS + 6], word 0x0000			; set read buffer position in memory (high word, 0 because we're addressing in 16bits)
	mov [EXT_DAP_POS + 8], dword 0x00000001		; set start block address (low dword, 2nd block)
	mov [EXT_DAP_POS + 12], dword 0x00000000	; set start block address (high dword)
	mov ah, 0x42								; command - extended read
	mov dl, [DRIVE_ID_POS]						; set drive ID
	mov si, EXT_DAP_POS							; set the buffer location where to store GPT header data from disk
	int	0x13									; call disk interrupt
	jnc test_gpt								; continiue if read went ok
	mov [ERR_MSG_POS], byte 0x34				; store error message
	jmp print_fault								; print a message and hang
	
test_gpt:										; Parse GPT signature (should be "EFI PART" - look up ASCII yourself)
	cmp dword [EXT_BUF_POS], 0x20494645			; should be "EFI " (or " IFE" in little-endian order)
	jne .test_gpt_sign_error					; jump to error if not even close
	cmp dword [EXT_BUF_POS + 4], 0x54524150		; should be "PART" (or "TRAP", pun intended?)
	jne .test_gpt_sign_error					; jump to error if not quite there
												; Test if we have any partitions at all
	cmp dword [EXT_BUF_POS + PART_COUNT_OFF], 0x00000001
	jl .test_gpt_count_error					; 0 partitions, it's a dead space
	cmp dword [EXT_BUF_POS + PART_COUNT_OFF], 0x00000004
	jg .set_part_count_4						; more than 4 partitions, we'll need only 4
	mov word cx, [EXT_BUF_POS + PART_COUNT_OFF]	; lees than or equal to 4 we read how much there is
	jmp read_part_arr							; jump to read
.set_part_count_4:
	mov word cx, 0x0004							; we only read first 4 partitions
	jmp read_part_arr							; jump to read
.test_gpt_sign_error:
	mov [ERR_MSG_POS], byte 0x35				; store error message
	jmp print_fault								; print a message and hang
.test_gpt_count_error:
	mov [ERR_MSG_POS], byte 0x36				; store error message
	jmp print_fault								; print a message and hang

read_part_arr:									; Read Partition array
	mov [EXT_DAP_POS + 8], dword 0x00000002		; set start block address (low dword, 3d block)
	mov [EXT_DAP_POS + 12], dword 0x00000000	; set start block address (high dword)
	mov ah, 0x42								; command - extended read
	mov dl, [DRIVE_ID_POS]						; set drive ID
	mov si, EXT_DAP_POS							; set the buffer location where to store GPT header data from disk
	int	0x13									; call disk interrupt
	jnc test_part								; continiue if read went OK
	mov [ERR_MSG_POS], byte 0x37				; store error message
	jmp print_fault								; print a message and hang

test_part:										; Read GPT partition entries to find if it's BBP (BIOS Boot Partition)
												; we only accept first partition to be BIOS bootable
	mov ax, word 0x0004							; set AX to 4 (we need it to calculate partition number)
	sub ax, cx									; subtract current step from AX (thus partition number is 4-4=0, 4-3=1, etc)
	mov dx, word 0x0080							; set DX to 128 bytes (partition entry size)
	mul dx										; multiply our result in AX to get the offset
	mov bx, ax									; store offset in BX (I think it's base offset register, is it?)
	cmp dword [EXT_BUF_POS + bx], 0x21686148	; and we test it, and it's uggly (GUID and little-endianness I mean)
	jne test_next								; stop, you're no BBP
	cmp dword [EXT_BUF_POS + bx + 4], 0x6E6F6449
	jne test_next								; nice try
	cmp dword [EXT_BUF_POS + bx + 8], 0x65654E74
	jne test_next								; yeah, almost got me
	cmp dword [EXT_BUF_POS + bx + 12], 0x49464564
	jne test_next								; sooo close
	jmp prepare_copy							; found it!

test_next:
	dec cx										; one step back means one step forward in partition entry array
	cmp cx, word 0x0000							; check if we're not at the end of our 4 entry array
	jg test_part								; try our luck on the next one
	mov [ERR_MSG_POS], byte 0x38				; store error message
	jmp print_fault								; print a message and hang

prepare_copy:									; Prepare our BBP copy
	mov [EXT_DAP_POS + 4], word ORG_POS			; set read buffer position in memory (low word)
	mov [EXT_DAP_POS + 6], word 0x0000			; set read buffer position in memory (high word, 0 because we're addressing in 16bits)
	mov ax, word [EXT_BUF_POS + bx + 40]		; store low word of end LBA
	sub ax, word [EXT_BUF_POS + bx + 32]		; subtract low word of start LBA
	cmp ax, 0x007F								; test if not bigger than 481Kib (962 secotrs)
	jg .copy_962
	mov [EXT_DAP_POS + 2], ax					; set number of blocks to transfer
	jmp .copy_lba_start
.copy_962:
	mov [EXT_DAP_POS + 2], word 0x007F			; set only 962 blocks 
	
.copy_lba_start:						
	mov ax, bx									; copy the start LBA to DAP
	add ax, 0x0020
	add ax, EXT_BUF_POS
	xor esi, esi								; clear esi, because bochs showed some dirt on in the high word :)
	mov si, ax									; set source index (partition entry start LBA)
	mov di, EXT_DAP_POS + 8						; set destination index (DAP start LBA)
	mov cx,	2									; how many blocks to copy (8 bytes)
	
	cld											; clear direction flag for copy
	rep											; repeat the copy cx times
	movsd										; move single dword
		
mbr_relocate:									; Reallocate us in memory, so that we have more contiguous area where to put 
												; our bootloader	
	mov si, bootstrap_start						; set source index (we only need the part from read_boot, rest is already a history)
	mov di, RELOC_POS							; set destination index (where to put our MBR)
	mov cx,	(bootstrap_end - bootstrap_start)	; how many blocks to copy (only from read_boot till padding)
	
	cld											; clear direction flag for copy
	rep											; repeat the copy cx times
	movsb										; move single byte
	
	; housekeeping
	xor eax, eax
	xor ebx, ebx
	xor ecx, ecx
	xor edx, edx
	xor si, si
	xor di, di

reset_disk_again:
	mov ah, 0x00								; command - reset disk
	mov dl, [DRIVE_ID_POS]						; set drive ID
	int 0x13									; call disk interrupt
	jnc RELOC_POS								; continiue to bootstrap if ok
	mov [ERR_MSG_POS], byte 0x39				; store error message
	jmp print_fault								; print a message and hang

bootstrap_start:								; Read bootable partition
	mov ah, 0x42								; command - extended read
	mov dl, [DRIVE_ID_POS]						; set drive ID
	mov si, EXT_DAP_POS							; set the buffer location where to store GPT header data from disk
	int	0x13									; call disk interrupt
	jc $										; halt on error
	jmp 0x0000:ORG_POS							; pass control over to BBP code (FUGLY!)
bootstrap_end:

print_fault:
	mov ah, 0x00								; command: set video mode (only now!)
	mov al, 0x03								; 3rd mode = 80x25 chars text screen
	int 0x10									; call video interrupt
	mov si, msg_err_title						; move error title into si
	jmp print_fault_title						; print title

print_char:
	mov ah, 0x0E								; command: print char
	mov bx, 0x01								; page and color?
	int 0x10									; call video interrupt

print_fault_title:
	lodsb										; lodsb loads ds:esi into al
    cmp al, 0x0
    jnz print_char								; print a character if there is any left

print_fault_code:
	mov al, byte [ERR_MSG_POS]					; move error code into si
	mov ah, 0x0E								; command: print char
	mov bx, 0x01								; page and color?
	int 0x10									; call video interrupt

halt:
	jmp halt									; halt!

msg_err_title:
	db "Error: ", 0

padding:										; Zerofill up to 510 bytes
	times 0x0200 - 2 - ($ - $$) db 0

magic:											; Add "magic number"
	dw 0xAA55									; Magic number - boot sector signature
